#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "Lookup.h"

int reverseComplement(int kmer, int k, int n);
void numToString(const int num, const int k, const int n, char *str);
void neighbours(int seq, int k, int m, int nNeighBases, int maxnneighs, int *nneighs, int *neighbours);

void reversecompstring(char *str, int len) {
    char *newstr = malloc( (len+1)*sizeof(char) );
    for (int i=0; i<len; i++) {
        char c = str[i]; 
        char cc;
        switch (c) {
            case 'A': cc='T'; break;
            case 'C': cc='G'; break;
            case 'G': cc='C'; break;
            case 'T': cc='A'; break;
        }
        newstr[len-i-1] = cc;
    }
    newstr[len] = '\0';
    strcpy(str, newstr);
    free(newstr);
    return;
}

int main(int argc, char **argv) {
    const int k=4;
    const int m=9;
    int nseq = (int)pow((double)k,(double)m);

    for (int i=0; i<nseq; i++) {
        int rc = reverseComplement(i, k, m);
        char seq  [k+1];
        char rcseq[k+1];
        numToString(i,  k, m, seq);
        numToString(rc, k, m, rcseq);
        seq  [m] = '\0';
        rcseq[m] = '\0';
        reversecompstring(seq, m);
        assert(strcmp(seq, rcseq) == 0);
    }

    const int nproc=32;
    int *lookup = generateLookupTable(k, m, nproc);

    srand(1);
    const int ntrials=20000;
    int balance[nproc];

    for (int p=0; p<nproc; p++) balance[p] = 0;

    for (int i=0; i<ntrials; i++) {
        int seq = rand() % nseq;
        int proc = lookup[seq];
        assert(proc < nproc);
        balance[proc]++;
    }

    int lo=balance[0];
    int hi=balance[0];
    for (int p=0; p<nproc; p++) {
        if (balance[p] < lo) lo=balance[p];
        if (balance[p] > hi) hi=balance[p];
        // printf("%d\n", balance[p]);
    }
    printf("\n");
    printf("Min %d, max %d, load balance = %f\n", lo, hi, 1.*lo/hi);

    free(lookup);
    
    {
        int nneighbases=3;
        int nshortseq = (int)pow((double)k,(double)(m-nneighbases));
        int seq = rand() % nshortseq;
        char origseqstr[m-nneighbases+1];
        char newseqstr [m+1];
        const int maxneighbours=4*4*4*(3+1);
        int nneighbours;
        int neighs[maxneighbours];
        neighbours(seq, k, m-nneighbases, nneighbases, maxneighbours, &nneighbours, neighs);

        numToString(seq, k, m-nneighbases, origseqstr); 
        printf("Nneighbours = %d\n", nneighbours);
        printf("Sequence     = <%s>\n",origseqstr);
        for (int i=0; i<nneighbours; i++) {
            numToString(neighs[i], k, m, newseqstr); 
            printf("Neighbour %3d: <%s> = %d\n", i, newseqstr, neighs[i] );
        }

        /* neighbours with nneighbases == 0 should just be self */
        int nseq = (int)pow((double)k,(double)(m));
        seq = rand() % nseq;
        numToString(seq, k, m, newseqstr); 
        printf("Sequence     = <%s>\n",newseqstr);
        neighbours(seq, k, m, 0, maxneighbours, &nneighbours, neighs);
        for (int i=0; i<nneighbours; i++) {
            numToString(neighs[i], k, m, newseqstr);
            printf("Neighbour %3d: <%s> = %d\n", i, newseqstr, neighs[i] );
        }

    
    }
    return 0;
}
